using System;
using System.IO;
using System.Linq;
using System.Collections.Generic;
using System.Runtime.CompilerServices;

namespace Model.Cartography.Nav
{
    public class NavigationGraph
    {
        public static readonly int INVALID = -1;

        private class TraversalState
        {
            public int Gen = INVALID;
            public Queue<int> Q = new Queue<int>();
            public int[] Dst, Par; /* Dst is not needed at the moment, but let's keep it for now */

            public int Next(int from) {
                return Par[from];
            }

            public override string ToString() {
                return $"TraversalState {{ Generation = {Gen} }}";
            }
        }

        private struct Route
        {
            public int At, To;

            public override string ToString() {
                return $"Route {{ At = {At}, To = {To} }}";
            }
        }

        private class VertexData
        {
            public List<int> Neighbors = new List<int>();
            public TraversalState TraversalFrom;

            public override string ToString() {
                return $"VertexData {{ Neighbors = {{ list of {Neighbors.Count} }}, TraversalState = { (TraversalFrom != null ? "exists" : "null") } }}";
            }
        }

        private int NextID = 0, Generation = 0;
        private List<VertexData> vertices = new List<VertexData>();
        private List<Route> routes = new List<Route>();

        public int VertexCount { get => NextID; }

#if DEBUG
        public StreamWriter DebugStream { get; set; }
        /* hacks on hacks all the way */

        private void debug(string s, [CallerMemberName] string caller = null)
            => DebugStream?.WriteLine((caller != null ? caller + ": " : "") + s);
#else
        private void debug(string s) { }
#endif
        public VertexHandle AddVertex() {
            debug($"vertex added with id={NextID} | Generation = {Generation}");
            return new VertexHandle(NextID++);
        }

        public void AddEdge(VertexHandle a, VertexHandle b)
        {
            ValidateVertexHandle(a);
            ValidateVertexHandle(b);

            EnsureStorage(a.ID);
            EnsureStorage(b.ID);

            vertices[a.ID].Neighbors.Add(b.ID);
            vertices[b.ID].Neighbors.Add(a.ID);

            Generation += 1;

            debug($"edge added between {a.ID} and {b.ID} | Generation = {Generation}");
        }

        public RouteHandle PlanRoute(VertexHandle from, VertexHandle to) {
            ValidateVertexHandle(from);
            ValidateVertexHandle(to);

            routes.Add(new Route() {
                At = from.ID,
                To = to.ID,
            });

            debug($"route added with id={routes.Count - 1}: { routes.Last() }");

            return new RouteHandle(routes.Count - 1);
        }

        public VertexHandle GetNextOnRoute(RouteHandle rh) {
            ValidateRouteHandle(rh);

            Route r = routes[rh.ID];

            debug($"get next on route: {r}");

            if (r.At == INVALID)
                return VertexHandle.INVALID;

            EnsureStorage(r.To);

            RecalcIfNecessary(rh);

            if (r.At == r.To) {
                // FIXME: do something
            }

            routes[rh.ID] = new Route() {
                At = vertices[r.To].TraversalFrom.Next(r.At),
                To = r.To,
            };

            return new VertexHandle(routes[rh.ID].At);
        }

        private void RecalcIfNecessary(RouteHandle rh) {
            Route r = routes[rh.ID];
            TraversalState tr = vertices[r.To].TraversalFrom;

            if (tr == null)
                tr = new TraversalState();

            debug($"tr = {tr}");

            if (tr.Gen != Generation) {
                tr.Q.Clear();
                tr.Q.Enqueue(r.To);
                tr.Dst = Enumerable.Range(0, VertexCount).Select(_ => INVALID).ToArray<int>();
                tr.Dst[r.To] = 0;
                tr.Par = Enumerable.Range(0, VertexCount).Select(_ => INVALID).ToArray<int>();
                tr.Par[r.To] = r.To;

                debug("TraversalState invalidated");
            } else {
                debug("reusing TraversalState");
            }

            if (tr.Par[r.At] == INVALID) {
                DoTraversal(tr, r.At);
                debug($"traversal from {r.To}");
            }

            vertices[r.To].TraversalFrom = tr;
        }

        private void ValidateRouteHandle(RouteHandle rh) {
            if (rh.ID >= routes.Count)
                throw new ArgumentOutOfRangeException();
        }

        private void ValidateVertexHandle(VertexHandle v)
        {
            if (v.ID >= NextID)
                throw new ArgumentOutOfRangeException();
        }

        private void EnsureStorage(int v)
        {
            if (v >= vertices.Count) {
                vertices.AddRange(Enumerable.Range(0, v - vertices.Count + 1)
                        .Select<int, VertexData>(_ => null));
            }

            if (vertices[v] == null)
                vertices[v] = new VertexData();
        }

        private void DoTraversal(TraversalState tr, int to) {
            var q = tr.Q;
            var dst = tr.Dst;
            var par = tr.Par;

            bool found = false;

            while (q.Count > 0 && !found) {
                int u = q.Dequeue();

                debug($"u = {u}");

                foreach (int v in vertices[u].Neighbors) {
                    debug($" v = {v}");
                    if (par[v] != INVALID)
                        continue;
                    debug($" v = {v}");

                    q.Enqueue(v);
                    dst[v] = dst[u] + 1;
                    par[v] = u;

                    if (v == to)
                        found = true;
                }
            }

            tr.Gen = Generation;
        }
    }
}
